# test_tower.py
#
# Copyright (c) 2025 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import pytest

from wretched_tower.tower import EmptyTowerError, PerilLevel, Tower


@pytest.mark.parametrize(
    "dice_size,expected_results",
    [
        (6, [1, 2, 3, 4, 5, 6]),
        (10, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
        (8, [1, 2, 3, 4, 5, 6, 7, 8]),
    ],
)
def test_possible_die_values(dice_size, expected_results) -> None:
    tower = Tower(dice_size=dice_size)
    assert tower.possible_values == expected_results


@pytest.mark.parametrize("dice_size", [-1, 0, 1])
def test_invalid_dice_size(dice_size) -> None:
    with pytest.raises(ValueError):
        Tower(dice_size=dice_size)


@pytest.mark.parametrize("dice_amount", [-2, 104])
def test_invalid_dice_amount(dice_amount) -> None:
    with pytest.raises(ValueError):
        Tower(dice_amount=dice_amount)


@pytest.mark.parametrize(
    "dice_size,expected_dict",
    [
        (6, {1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0}),
        (10, {1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0, 10: 0}),
    ],
)
def test_generate_result_dict(dice_size, expected_dict) -> None:
    tower = Tower(dice_size=dice_size)
    assert tower.get_result_dict_template() == expected_dict


def test_tower_roll() -> None:
    expected_dice = 100
    tower = Tower()
    assert expected_dice == tower.get_dice_left()
    assert len(tower.roll_distributions) == 0
    tower.roll_tower()
    assert len(tower.roll_distributions) == 1
    assert expected_dice == tower.roll_distributions[0].dice_rolled
    assert (
        tower.get_dice_left()
        == expected_dice - tower.roll_distributions[0].dice_results[1]
    )
    expected_dice = tower.get_dice_left()
    tower.roll_tower()
    assert len(tower.roll_distributions) == 2
    assert expected_dice == tower.roll_distributions[1].dice_rolled
    assert (
        tower.get_dice_left()
        == expected_dice - tower.roll_distributions[1].dice_results[1]
    )
    for distribution in tower.roll_distributions:
        total_dice_rolled = 0
        for _key, value in distribution.dice_results.items():
            total_dice_rolled += value
        assert total_dice_rolled == distribution.dice_rolled


@pytest.mark.parametrize(
    "dice_amount,expected_peril",
    [
        (100, PerilLevel.HEALTHY),
        (75, PerilLevel.HEALTHY),
        (59, PerilLevel.WOUNDED),
        (28, PerilLevel.WOUNDED),
        (24, PerilLevel.MORTALITY),
    ],
)
def test_tower_peril_calculation(dice_amount, expected_peril) -> None:
    tower = Tower(dice_amount=dice_amount)
    assert tower.get_peril_level() == expected_peril


def test_tower_peril_dead() -> None:
    tower = Tower()
    # Use private access to this property to mimic the dead scenario.
    tower._dice_left = 0
    assert tower.get_peril_level() == PerilLevel.DEAD
    with pytest.raises(EmptyTowerError):
        tower.roll_tower()



# __init__.py
#
# Copyright (c) 2025 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

from wretched_tower.app import TowerApp

__version__ = "0.1.1"


def main() -> None:  # no cov
    tower_app = TowerApp()
    tower_app.run()



# tower.py
#
# Copyright (c) 2025 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import random
from enum import Enum
from typing import ClassVar


class EmptyTowerError(Exception):
    """Used when actions are taken on a tower where no dice remain."""
    pass


class PerilLevel(Enum):
    HEALTHY = 1
    WOUNDED = 2
    MORTALITY = 3
    DEAD = 4


@dataclasses.dataclass
class RollDistribution:
    """
    An object that references the number of dice rolled and dict
    representing the counts for each possible result.
    """

    dice_rolled: int
    dice_results: dict[int, int]


@dataclasses.dataclass
class RollResult:
    """An object that contains the overall result of a tower roll: rolled vs lost."""

    dice_rolled: int
    dice_lost: int


@dataclasses.dataclass
class Tower:
    """
    A given instance of the dice tower.

    Attributes:
        roll_distributions (list[dict[int,int]]): A list of the previous roll results
            and their counts.
        possible_values (list[int]): A list of the possible die values based on size.
    """

    roll_distributions: list[RollDistribution]
    possible_values: list[int]
    _dice_size: int = 6
    _dice_left: int = 100

    HEALTHY_LIMIT: ClassVar[int] = 60
    FATAL_LIMIT: ClassVar[int] = 25

    def __init__(self, dice_size: int = 6, dice_amount: int = 100) -> None:
        if dice_size < 2:  # noqa: PLR2004
            msg = "Dice must have more than one side!"
            raise ValueError(msg)
        self._dice_size = dice_size
        self.set_dice_left(dice_amount)
        self.possible_values = self._get_possible_die_values()
        self.roll_distributions = []

    def __str__(self) -> str:  # no cov
        return f"Tower ({self.get_dice_left()}"

    def get_dice_left(self) -> int:  # no cov
        """
        Get the dice remaining to be rolled.

        Returns:
            int: The dice remaining.
        """
        return self._dice_left

    def set_dice_left(self, dice_left: int) -> None:
        """
        Allows you to set the dice left value with some validations applied.
        Allowed values must be between 0 and 100.

        Args:
            dice_left (int): The dice remaining.

        Raises:
            ValueError: If the dice amount exceeds 100 or is less than 0.
        """
        if dice_left > 100:  # noqa: PLR2004
            msg = "Tower cannot exceed 100 dice!"
            raise ValueError(msg)
        elif dice_left < 0:
            msg = "Tower dice amount cannot be a negative number!"
            raise ValueError(msg)
        self._dice_left = dice_left

    def _get_possible_die_values(self) -> list[int]:
        """
        Get a list of all the possible values based on the dice size.

        Returns:
            list[int]: The possible values based on the die size.
        """
        current_value = 1
        possible_values = []
        while current_value <= self._dice_size:
            possible_values.append(current_value)
            current_value += 1
        return possible_values

    def get_result_dict_template(self) -> dict[int, int]:
        """Get a dictionary of possible dice values and zeroed counts."""
        results = {}
        for x in self.possible_values:
            results[x] = 0
        return results

    def roll_tower(self) -> RollResult:
        """
        Using the dice remaining, roll them and then remove any that are ones,
        recording the results.

        Returns:
            RollResult: The results of the roll as a RollResult object.
        """
        dice_to_roll = self.get_dice_left()
        if dice_to_roll == 0:
            msg = "Tower does not have any dice remaining."
            raise EmptyTowerError(msg)
        results = self.get_result_dict_template()
        for x in self.possible_values:
            results[x] = 0
        for _x in range(self._dice_left):
            die_result = random.randint(1, self._dice_size)  # nosec  # noqa: S311
            results[die_result] += 1
        self.set_dice_left(self._dice_left - results[1])
        self.roll_distributions.append(
            RollDistribution(dice_rolled=dice_to_roll, dice_results=results)
        )
        return RollResult(dice_rolled=dice_to_roll, dice_lost=results[1])

    def get_peril_level(self) -> PerilLevel:
        if self._dice_left == 0:
            return PerilLevel.DEAD
        elif self._dice_left < self.FATAL_LIMIT:
            return PerilLevel.MORTALITY
        elif self._dice_left < self.HEALTHY_LIMIT:
            return PerilLevel.WOUNDED
        else:
            return PerilLevel.HEALTHY



# app.py
#
# Copyright (c) 2025 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

from textual.app import App, ComposeResult
from textual.color import Color
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import DataTable, Digits, Footer, Header, Static

from wretched_tower.tower import EmptyTowerError, PerilLevel, Tower


def get_dice_display(dice_left: int) -> str:
    return str(dice_left) if dice_left > 0 else "☠️ 0 ☠️"


def get_tower_color_from_peril_level(peril_level: PerilLevel) -> Color:
    match peril_level:
        case PerilLevel.MORTALITY:
            return Color.parse("orange")
        case PerilLevel.WOUNDED:
            return Color.parse("yellow")
        case PerilLevel.DEAD:
            return Color.parse("red")
        case PerilLevel.HEALTHY:
            return Color.parse("green")
        case _:  # no cov
            return Color.parse("green")


def get_peril_meter_display_from_peril_level(peril_level: PerilLevel) -> str:
    match peril_level:
        case PerilLevel.MORTALITY:
            return "MORTAL DANGER"
        case PerilLevel.WOUNDED:
            return "WOUNDED"
        case PerilLevel.DEAD:
            return "DEAD"
        case PerilLevel.HEALTHY:
            return "HEALTHY"
        case _:  # no cov
            return "HEALTHY"


class DiceCounter(Digits):
    pass


class PerilMeter(Static):
    pass


class TowerStatus(Widget):
    """A widget that displays the current status of the tower."""

    peril_level = reactive(PerilLevel.HEALTHY)
    tower_color = reactive(Color.parse("green"))

    def __init__(self, tower: Tower, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        super().__init__()
        self.peril_level = tower.get_peril_level()
        self.tower_color = get_tower_color_from_peril_level(self.peril_level)

    def compose(self) -> ComposeResult:
        yield Static("[b]Dice Remaining[/b]")
        yield DiceCounter("100", id="dice")
        yield PerilMeter(get_peril_meter_display_from_peril_level(self.peril_level))
        yield DataTable(name="Roll Results")

    def on_mount(self) -> None:
        table = self.query_one(DataTable)
        table.add_columns("Rolled", "1", "2", "3", "4", "5", "6")

    def watch_tower_color(self, color: Color) -> None:
        self.styles.color = color

    def watch_peril_level(self, peril_level: PerilLevel) -> None:
        self.tower_color = get_tower_color_from_peril_level(peril_level)


class TowerApp(App):
    """A TUI that manages a tumbling tower mechanic via die rolls."""

    BINDINGS = [
        ("d", "toggle_dark", "Toggle dark mode"),
        ("r", "roll_tower", "Roll tower dice"),
        ("ctrl+n", "new_tower", "Start new tower"),
    ]
    TITLE = "Wretched Tower"
    # theme = "tokyo-night"
    tower = Tower()

    CSS = """
        TowerStatus Static {
            text-align: center;
        }
        DiceCounter {
            text-align: center;
        }
    """

    def compose(self) -> ComposeResult:
        """Create child widgets"""
        yield Header()
        yield TowerStatus(tower=self.tower, id="status")
        yield Footer()

    def action_toggle_dark(self) -> None:
        """Toggles between dark and light mode."""
        self.theme = (  # type: ignore
            "textual-dark" if self.theme == "textual-light" else "textual-light"
        )

    def action_roll_tower(self) -> None:
        """Roll the tower and send the results to the child widgets as needed."""
        try:
            self.tower.roll_tower()
            dice_left = self.tower.get_dice_left()
            tower_status_widget = self.query_one(TowerStatus)
            tower_status_widget.peril_level = self.tower.get_peril_level()
            self.query_one(DiceCounter).update(get_dice_display(dice_left))
            self.query_one(PerilMeter).update(
                get_peril_meter_display_from_peril_level(
                    tower_status_widget.peril_level
                )
            )
            last_result = self.tower.roll_distributions[-1]
            tower_status_widget.query_one(DataTable).add_row(
                last_result.dice_rolled,
                last_result.dice_results[1],
                last_result.dice_results[2],
                last_result.dice_results[3],
                last_result.dice_results[4],
                last_result.dice_results[5],
                last_result.dice_results[6],
            )
            if dice_left == 0:
                self.notify("[b][red]You Have Died[/red][/b]")
        except EmptyTowerError:
            self.notify("You are already dead.", severity="error")

    def action_new_tower(self) -> None:
        self.tower = Tower()
        dice_left = self.tower.get_dice_left()
        tower_status_widget = self.query_one(TowerStatus)
        self.query_one(DiceCounter).update(get_dice_display(dice_left))
        tower_status_widget.peril_level = self.tower.get_peril_level()
        self.query_one(PerilMeter).update(
            get_peril_meter_display_from_peril_level(tower_status_widget.peril_level)
        )
        tower_status_widget.query_one(DataTable).clear()



# test_app.py
#
# Copyright (c) 2025 Daniel Andrlik
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import pytest
from textual.color import Color

from wretched_tower.app import (
    DiceCounter,
    PerilMeter,
    TowerApp,
    TowerStatus,
    get_dice_display,
    get_peril_meter_display_from_peril_level,
    get_tower_color_from_peril_level,
)
from wretched_tower.tower import PerilLevel


@pytest.mark.parametrize(
    "dice_value,expected_output",
    [(100, "100"), (75, "75"), (23, "23"), (0, "☠️ 0 ☠️"), (-2, "☠️ 0 ☠️")],
)
def test_get_dice_display(dice_value, expected_output):
    assert get_dice_display(dice_value) == expected_output


@pytest.mark.parametrize(
    "peril_level,expected_color",
    [
        (PerilLevel.HEALTHY, Color.parse("green")),
        (PerilLevel.WOUNDED, Color.parse("yellow")),
        (PerilLevel.MORTALITY, Color.parse("orange")),
        (PerilLevel.DEAD, Color.parse("red")),
    ],
)
def test_get_tower_color_from_peril_level(peril_level, expected_color):
    assert get_tower_color_from_peril_level(peril_level) == expected_color


@pytest.mark.parametrize(
    "peril_level,expected_output",
    [
        (PerilLevel.HEALTHY, "HEALTHY"),
        (PerilLevel.WOUNDED, "WOUNDED"),
        (PerilLevel.MORTALITY, "MORTAL DANGER"),
        (PerilLevel.DEAD, "DEAD"),
    ],
)
def test_get_peril_meter_display_from_peril_level(peril_level, expected_output):
    assert get_peril_meter_display_from_peril_level(peril_level) == expected_output


def test_app_layout(snap_compare) -> None:
    assert snap_compare("../src/wretched_tower/app.py")


@pytest.mark.asyncio
async def test_dark_mode_toggle() -> None:
    app = TowerApp()
    async with app.run_test() as pilot:
        await pilot.press("d")
        assert app.theme == "textual-light"
        await pilot.press("d")
        assert app.theme == "textual-dark"


@pytest.mark.asyncio
async def test_roll_display() -> None:
    app = TowerApp()
    async with app.run_test() as pilot:
        assert pilot.app.query_one(TowerStatus).styles.color == Color.parse("green")
        while pilot.app.tower.get_dice_left() > 0:
            await pilot.press("r")
            await pilot.pause()
            dice_left = pilot.app.tower.get_dice_left()
            if dice_left == 0:
                assert len(pilot.app._notifications) == 1
            assert pilot.app.query_one(DiceCounter).value == get_dice_display(
                dice_left
            )
            displayed_color = pilot.app.query_one(TowerStatus).styles.color
            tower_status = app.query_one(TowerStatus)
            assert tower_status.peril_level == pilot.app.tower.get_peril_level()
            assert pilot.app.query_one(
                PerilMeter
            )._content == get_peril_meter_display_from_peril_level(
                pilot.app.tower.get_peril_level()
            )
            match pilot.app.tower.get_peril_level():
                case PerilLevel.HEALTHY:
                    assert displayed_color == Color.parse("green")
                case PerilLevel.WOUNDED:
                    assert displayed_color == Color.parse("yellow")
                case PerilLevel.MORTALITY:
                    assert displayed_color == Color.parse("orange")
                case PerilLevel.DEAD:
                    assert displayed_color == Color.parse("red")


@pytest.mark.asyncio
async def test_new_tower() -> None:
    app = TowerApp()
    async with app.run_test() as pilot:
        pilot.app.tower._dice_left = 80
        await pilot.press("r")
        await pilot.pause()
        assert pilot.app.query_one(DiceCounter).value != "100"
        await pilot.press("ctrl+n")
        await pilot.pause()
        assert pilot.app.query_one(DiceCounter).value == "100"


@pytest.mark.asyncio
async def test_already_dead_notification() -> None:
    app = TowerApp()
    async with app.run_test() as pilot:
        app.tower._dice_left = 0
        assert len(pilot.app._notifications) == 0
        await pilot.press("r")
        await pilot.pause()
        assert len(pilot.app._notifications) == 1
